/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "config.h"

#include <sys/stat.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h> /* time(), localtime() */
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stropts.h>
#include <poll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <libgen.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/un.h> /* needed by create_pf_socket */
#include <malloc.h> /* mallinfo */

/* some internal constants */
#define STRINGSIZE 4096
#define MAXLINE 255


#define MIN(a,b) (a) < (b) ? (a) : (b)
#define MAX(a,b) (a) > (b) ? (a) : (b)
#define ABS(a) (a) > 0 ? (a) : -(a)

/* Loglevels */
#define DEBUG 1
#define INFO  2
#define NOTICE 3
#define WARNING 4
#define ERROR 5
#define CRIT 6
#define EMERG 7


#define COMPILE_TIME_LOGLEVEL 0
#if DEBUG > COMPILE_TIME_LOGLEVEL
#define PARA_DEBUG_LOG(f,...) para_log(DEBUG, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_DEBUG_LOG(...)
#endif

#if INFO > COMPILE_TIME_LOGLEVEL
#define PARA_INFO_LOG(f,...) para_log(INFO, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_INFO_LOG(...)
#endif

#if NOTICE > COMPILE_TIME_LOGLEVEL
#define PARA_NOTICE_LOG(f,...) para_log(NOTICE, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_NOTICE_LOG(...)
#endif

#if WARNING > COMPILE_TIME_LOGLEVEL
#define PARA_WARNING_LOG(f,...) para_log(WARNING, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_WARNING_LOG(...)
#endif

#if ERROR > COMPILE_TIME_LOGLEVEL
#define PARA_ERROR_LOG(f,...) para_log(ERROR, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_ERROR_LOG(...)
#endif

#if CRIT > COMPILE_TIME_LOGLEVEL
#define PARA_CRIT_LOG(f,...) para_log(CRIT, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_CRIT_LOG(...)
#endif

#if EMERG > COMPILE_TIME_LOGLEVEL
#define PARA_EMERG_LOG(f,...) para_log(EMERG, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_EMERG_LOG(...)
#endif



/* used in several contexts */
#define SET 0
#define GET 1
#define INCR 2
#define RESET 3


/* These are the permission flags that can be set individually
 * for any command
 */
#define DB_READ 1
#define DB_WRITE 2
#define AFS_READ 4
#define AFS_WRITE 8

#define COPYRIGHT "Copyright (c) 1997-2005 by Andre Noll"

#define LICENSE "This is free software with ABSOLUTELY NO WARRANTY. " \
	"See COPYING for details."

#define EMAIL "Send bug reports and " \
	"questions to <maan@systemlinux.org>."

#define AWAITING_DATA_MSG "\nAwaiting Data."
#define PROCEED_MSG "\nProceed.\n"
#define PROCEED_MSG_LEN strlen(PROCEED_MSG)
#define EOC_MSG "\nEnd of Command."
#define CHALLENGE_RESPONSE_MSG "challenge_response:"

struct gui_command {
	char *key;
	char *name;
	char *description;
	void (*handler)(int);
};

/* net */
#define UNIX_PATH_MAX 108
struct hostent *get_host_info(char *);
int get_socket(void);
void init_sockaddr(struct sockaddr_in*, int, struct hostent*);
int do_connect(int, struct sockaddr_in *);
int send_buffer(int, char *);
int send_bin_buffer(int, const char *, size_t);
int send_va_buffer(int, char *, ...);
int recv_buffer(int, char *, ssize_t);
int recv_bin_buffer(int, char *, ssize_t);
int do_accept(int, struct sockaddr_in *);
int create_pf_socket(const char *, struct sockaddr_un *, int mod);
int init_unix_addr(struct sockaddr_un *, const char *);
int recv_cred_buffer(int, char *, size_t, struct ucred *);
ssize_t send_cred_buffer(int, char*);
char *para_hostname(void);
int init_tcp_socket(int port);
int recv_pattern(int fd, const char *pattern, size_t bufsize);

/* gui */
int file_exists(const char *);
int para_open_audiod_pipe(char *);

/* exec */
int para_open_fifo(const char*, const int);
int para_exec(pid_t *, const char *, char *const [], int *);
int para_exec_cmdline_pid(pid_t *, char *, int *);

/* strings */
void chop(char*);
void *para_realloc(void*, size_t);
void *para_malloc(size_t);
void *para_calloc(size_t size);
char *para_strdup(const char *);
char *para_basename(const char *name);
char *para_dirname(const char *name);
char *make_message(const char*, ...) __attribute__((format (printf, 1, 2)));
char *para_strcat(char *, const char *);
int para_fread(void *, size_t, size_t, FILE *);
char *para_tmpname(void);
int para_mkstemp(char *, mode_t);
struct regex_list {
	char *regex;
	char *replacement;
};
char *s_a_r(const char *, const char *, const char *);
char *s_a_r_list(struct regex_list *, char *);
char *para_logname(void);
char *para_homedir(void);
int split_args(char *, char ***, int);

/* signal */
int para_signal_init(int *spipe);
int para_install_sighandler(int);
void para_reap_children(void);
pid_t para_reap_child(void);
int para_next_signal(void);

/* time */

int tv_diff(const struct timeval *b, const struct timeval *a, struct timeval *diff);
long unsigned tv2ms(const struct timeval*);
void d2tv(double, struct timeval*);
void tv_add(const struct timeval*, const struct timeval *, struct timeval *);
void tv_scale(const unsigned long, const struct timeval *, struct timeval *);
void tv_divide(const unsigned long div, const struct timeval *tv,
	struct timeval *result);
int tv_convex_combination(const long a, const struct timeval *tv1,
		const long b, const struct timeval *tv2,
		struct timeval *result);
void ms2tv(const long unsigned n, struct timeval *tv);

/* daemon */
void daemon_init(void);
FILE *open_log(char *);
void close_log(FILE *);
void log_welcome(const char *, int);
void para_drop_privileges(const char *username);
int server_uptime(int set_or_get);
char *uptime_str(int i);

/* stat */
int read_audiod_pipe(int, void (*)(char *));
enum {
	SI_STATUS_BAR,		SI_STATUS,		SI_NUM_PLAYED,
	SI_MTIME,		SI_LENGTH_MIN,		SI_LENGTH_SEC,
	SI_FILE_SIZE,		SI_STATUS_FLAGS,	SI_FORMAT,
	SI_SCORE,		SI_AUDIO_INFO1,		SI_AUDIO_INFO2,
	SI_AUDIO_INFO3,		SI_DBINFO1,		SI_DBINFO2,
	SI_DBINFO3,		SI_DECODER_FLAGS,	SI_AUDIOD_STATUS,
	SI_PLAY_TIME,		SI_UPTIME,		SI_OFFSET,
	SI_LENGTH,		SI_STREAM_START,	SI_CURRENT_TIME,
	SI_AUDIOD_UPTIME,	SI_DBTOOL,
};
#define NUM_STAT_ITEMS (SI_DBTOOL + 1)
int stat_line_valid(const char *);
void dump_stat_line(char *);
int stat_client_add(int);
void dump_empty_status(void);
int check_buf_for_items(char *, int, void (*)(char *), int);

struct stat_item_data {
	char *prefix, *postfix;
	unsigned x, y, len;
	int fg, bg, align;
};

/* gui_theme */
struct gui_theme {
	char *name;
	char *author;
	int sb_fg, sb_bg;
	int cmd_fg, cmd_bg;
	int output_fg, output_bg;
	int msg_fg, msg_bg;
	int err_msg_fg, err_msg_bg;
	int welcome_fg, welcome_bg;
	int sep_fg, sep_bg;
	char *sep_str;
	int default_fg, default_bg;

	int top_lines_default, top_lines_min;
	int lines_min, cols_min;
	struct stat_item_data data[NUM_STAT_ITEMS];
};

void init_theme(int i, struct gui_theme *);
void next_theme(struct gui_theme *);
void prev_theme(struct gui_theme *);
#define LEFT 1
#define RIGHT 2
#define CENTER 3


void para_log(int, char*, ...) __attribute__((format (printf, 2, 3)));

/* taken from printf man page */
#define PARA_VSPRINTF(fmt, p) \
{ \
	int n, size = 100; \
	p = para_malloc(size); \
	while (1) { \
		va_list ap; \
		/* Try to print in the allocated space. */ \
		va_start(ap, fmt); \
		n = vsnprintf(p, size, fmt, ap); \
		va_end(ap); \
		/* If that worked, return the string. */ \
		if (n > -1 && n < size) \
			break; \
		/* Else try again with more space. */ \
		if (n > -1) /* glibc 2.1 */ \
			size = n + 1; /* precisely what is needed */ \
		else /* glibc 2.0 */ \
			size *= 2; /* twice the old size */ \
		p = para_realloc(p, size); \
	} \
}
